package profiler;

import ij.IJ;
import ij.ImagePlus;
import ij.gui.OvalRoi;
import ij.gui.Overlay;
import ij.gui.PointRoi;
import ij.gui.Roi;
import ij.io.FileInfo;
import ij.plugin.PlugIn;
import java.awt.Color;
import java.awt.Font;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import javax.swing.JOptionPane;
import loci.common.DebugTools;
import shared.GUV;
import shared.GUVReader;
import shared.GUVSet;
import shared.IJLogger;
import shared.Profile;

/**Main model*/
public class Profiler implements PlugIn {

   public static final String TITLE = "GUV Profiler";
   static final String VERSION = "2.4";

   final GUVTableModel guvTableModel;
   final ProfileTableModel profileTableModel;

   ImagePlus imp;
   List<GUVSet> list;

   private final PropertyChangeSupport pcs;

   private GUVSet set;
   private GUV activeGUV;
   private int channel, slice, frame, transmittedChannel;

   public Profiler() {
      IJLogger logger = new IJLogger(TITLE);
      //Redirect all output to the IDE console if running within an IDE
      String logging = System.getenv("LOGGING");
      if (logging != null && logging.equals("system")) {
         logger.setSystemLogging();
      }
      pcs = new PropertyChangeSupport(this);
      guvTableModel = new GUVTableModel(this);
      profileTableModel = new ProfileTableModel(guvTableModel);
   }

   /**Plugin checks: Fiji legacy "Lasso" tool, circle fitter and MaxInscribedCircles.*/
   public static void checkPlugins() {
      try {
         Class.forName("loci.plugins.BF");
         DebugTools.setRootLevel("ERROR");
      } catch (Exception e) {
         JOptionPane.showMessageDialog(null,
               "Bio-Formats plugin not found.\n" +
                     "It should be built-in in Fiji.", TITLE, JOptionPane.ERROR_MESSAGE);
         return;
      }
      try {
         Class.forName("plugin.Lasso");
      } catch (ClassNotFoundException e) {
         JOptionPane.showMessageDialog(null,
               "Required Blow/Lasso tool not found.\n" +
                     "It should be built-in in Fiji.", TITLE, JOptionPane.ERROR_MESSAGE);
         return;
      }
      try {
         Class.forName("fiji.util.Circle_Fitter");
      } catch (ClassNotFoundException e) {
         JOptionPane.showMessageDialog(null,
               "Required Circle Fitter plugin not found.\n" +
                     "It should be built-in in Fiji.", TITLE, JOptionPane.ERROR_MESSAGE);
         return;
      }
      try {
         Class.forName("ch.epfl.biop.MaxInscribedCircles");
      } catch (ClassNotFoundException e) {
         JOptionPane.showMessageDialog(null,
               "Required BIOP plugin not found.\n" +
                     "Install it with Help -> Update... -> Manage Update Sites ->\n" +
                     "Scroll down and check the 'PTBIOP' update site.",
               TITLE, JOptionPane.ERROR_MESSAGE);
      }
   }

   @Override
   public void run(String arg) {
      checkPlugins();
      GUV.readPreferences();
      Dialog dialog = new Dialog(this, TITLE, VERSION);
      dialog.setVisible(true);
   }

   /**Adjusts a line profile and re-calculates the statistics.
    * */
   void adjustLine(int guvIndex, int profileIndex) {
      if (set == null || imp == null) {
         return;
      }
      GUV guv = set.get(guvIndex);
      guv.profiles.get(profileIndex).adjustLine(imp, (PointRoi) imp.getRoi());
      guv.calculateAll();
      set.calculateAll();
      updateStatistics(imp.getC() - 1);
   }

   /**Adds a new GUV from the current selection of the current image.
    * @param displayOutlines indicates to draw the outline of the new GUV*/
   int add(boolean displayOutlines) {
      if (imp == null || set == null) {
         return -1;
      }
      Roi roi = imp.getRoi();
      if (roi == null) {
         return -1;
      }
      OvalRoi circle = GUVReader.fitCircle(imp, roi, 5, false);
      circle.setPosition(0, slice, frame);
      GUV guv = new GUV(set.getNextID(), imp, circle, slice, frame, set.all);
      guv.measureAll(imp);
      set.add(guv);
      set.calculateAll();
      updateStatistics(channel - 1);
      displayProfiles(set.count() - 1, displayOutlines);
      return set.count() - 1;
   }

   void addListener(PropertyChangeListener listener) {
      pcs.addPropertyChangeListener(listener);
   }

   /**Deletes the GUVs with the indices supplied.
    * @param index the indices*/
   void delete(int... index) {
      Overlay overlay = imp.getOverlay();
      for (int i : index) {
         GUV guv = set.delete(i);
         overlay.remove(guv.roi);
      }
      set.calculateAll();
      updateStatistics(channel - 1);
      imp.updateAndDraw();
   }

   /**Deletes all GUVs of the current GUV set.*/
   void deleteAll() {
      imp.getOverlay().clear();
      set.clear();
      imp.updateAndDraw();
   }

   /**Toggles the display of GUV outlines
    * @param show displays outlines if <code>true</code>*/
   void displayOutlines(boolean show) {
      Overlay overlay = imp.getOverlay();
      for (GUV guv : set.getAll()) {
         if (show) {
            overlay.add(guv.roi);
         } else {
            overlay.remove(guv.roi);
         }
      }
      imp.updateAndDraw();
   }

   /**Displays a single profile of a given GUV
    * @param index the index of the GUV
    * @param profile the index of the profile to display*/
   void displayProfile(int index, int profile) {
      Overlay overlay = imp.getOverlay();
      overlay.clear();
      if (index > -1) {
         activeGUV = set.get(index);
         Profile p = activeGUV.profiles.get(profile);
         overlay.add(p.line, String.valueOf(p.id));
      }
      imp.updateAndDraw();
   }

   /**Displays all used profiles of the GUV with the provided index. Removes displayed profiles
    * if the index is smaller than zero
    * @param index the GUV index
    * @param displayOutline <code>true</code> displays the circular ROi*/
   void displayProfiles(int index, boolean displayOutline) {
      if (imp == null) {
         return;
      }
      Overlay overlay = imp.getOverlay();
      overlay.clear();
      if (index > -1) {
         activeGUV = set.get(index);
         for (Profile p : activeGUV.profiles) {
            if (p.isUsed()) {
               overlay.add(p.line, String.valueOf(p.id));
            }
         }
         if (displayOutline) {
            overlay.add(activeGUV.roi);
         }
      } else {
         //Index is -1, the user deselected a row in the table
         displayOutlines(displayOutline);
      }
      imp.updateAndDraw();
   }

   /**Provides the GUV with the supplied index.
    * @param index the index*/
   @SuppressWarnings("unused")
   GUV get(int index) {
      if (set != null && index >= 0) {
         return set.get(index);
      }
      return null;
   }

   /**Provides the index of the current background GUV.
    * @return the index or -1 if there is none*/
   int getControl() {
      return set.getControlIndex();
   }

   /**Re-measures all profiles/GUVs in all channels at its current slice and frame.*/
   void measureAll() {
      if (set == null || imp == null) {
         return;
      }
      set.measureAllChannels(imp);
      updateStatistics(channel - 1);
   }

   /**Resets this profiler.
    * @param closeImage <code>true</code> if the image should be closed.*/
   void reset(boolean closeImage) {
      if (imp != null) {
         Overlay overlay = imp.getOverlay();
         if (overlay != null) {
            overlay.clear();
         }
         try {
            imp.updateAndDraw();
         } catch (Exception ignored) {}
         if (closeImage) {
            imp.changes = false;
            imp.close();
         }
      }
      imp = null;
      set = null;
      activeGUV = null;
      channel = 0;
      slice = 0;
      frame = 0;
      transmittedChannel = -1;
      pcs.firePropertyChange("mean", null, 0d);
      pcs.firePropertyChange("corrected", null, 0d);
      pcs.firePropertyChange("SD", null, 0d);
   }

   /**Sets the active image.
    * @param imp the image
    * @param displayOutlines displays all GUV oval ROIs*/
   void setImage(ImagePlus imp, boolean displayOutlines) {
      if (imp != null) {
         this.imp = imp;
      }
      String str = imp != null ? imp.getProp("transmitted") : null;
      if (str == null) {
         transmittedChannel = -1;
      } else {
         transmittedChannel = Integer.parseInt(str);
      }
      //noinspection ConstantConditions
      if (imp.getOverlay() == null) {
         imp.setOverlay(new Overlay());
      } else {
         imp.getOverlay().clear();
      }
      Overlay overlay = imp.getOverlay();
      overlay.drawLabels(true);
      overlay.drawNames(true);
      overlay.setLabelFont(new Font(Font.SANS_SERIF, Font.BOLD, 16));
      overlay.setLabelColor(Color.CYAN);
      overlay.setFillColor(Color.CYAN);
      //Try to find zipped ROIs or the mask image
      FileInfo info = imp.getOriginalFileInfo();
      if (info == null) {
         return;
      }
      Path dir = Paths.get(info.directory);
      if (Files.notExists(dir)) {
         JOptionPane.showMessageDialog(null, "Directory of the active image cannot be read:\n" +
               dir, TITLE, JOptionPane.ERROR_MESSAGE);
         return;
      }
      Path zip = dir.resolve(imp.getShortTitle()+".zip");
      if (Files.exists(zip) && !Files.isReadable(zip)) {
         return;
      }
      if (Files.notExists(zip)) {
         Path labels = dir.resolve(imp.getShortTitle()+"_cp_masks.tif");
         if (Files.exists(labels) && Files.isReadable(labels)) {
            ImagePlus masks = IJ.openImage(String.valueOf(labels));
            //The mask image may not resized to the original size
            if (masks.getWidth() != imp.getWidth()) {
               ImagePlus resized = masks.resize(imp.getWidth(), imp.getHeight(), "none");
               masks.changes = false;
               masks.close();
               masks = resized;
            }
            List<GUVSet> list = GUVReader.readLabels(imp, masks);
            masks.changes = false;
            masks.close();
            GUVReader.measure(list, imp);
            GUVReader.saveRoiSet(list, zip.toString());
         }
      }
      if (Files.exists(zip)) {
         list = GUVReader.readRoiSet(imp,zip.toString());
      } else {
         /*Either there is no ZIP because there are no ROIs, or we are entirely in manual mode.
         * Construct an empty list of GUV sets.*/
         list = new ArrayList<>(imp.getNFrames() * imp.getNSlices());
         for (int i = 1; i <= imp.getNFrames(); i++) {
            for (int j = 1; j <= imp.getNSlices(); j++) {
               GUVSet set = new GUVSet(i, j, imp.getNChannels());
               list.add(set);
            }
         }
      }
      updatePosition(imp.getC(), imp.getZ(), imp.getT());
      if (displayOutlines) {
         //noinspection ConstantConditions
         displayOutlines(displayOutlines);
      }
   }

   /**Sets a GUV as background/noise.*/
   void setAsControl(int index) {
      set.setAsControl(index);
      set.calculateAll();
      updateStatistics(channel - 1); //the channel is one-based as in ImagePlus
   }

   void updatePosition(int channel, int slice, int frame) {
      if (this.slice != slice || this.frame != frame) {
         this.slice = slice;
         this.frame = frame;
         for (GUVSet set : list) {
            if (set.frame == frame && set.slice == slice) {
               this.set = set;
               guvTableModel.set(set);
            }
         }
         updateChannel(channel);
      } else if (this.channel != channel) {
         updateChannel(channel);
      }
   }

   void updateStatistics(int channel) {
      pcs.firePropertyChange("mean", null, set.mean[channel]);
      pcs.firePropertyChange("corrected", null, getCorrectedMean(channel));
      pcs.firePropertyChange("SD", null, set.sd[channel]);
   }

   private double getCorrectedMean(int channel) {
      GUV guv = set.getControl();
      if (guv == null) {
         return set.mean[channel];
      } else {
         return set.mean[channel] - guv.mean[channel];
      }
   }

   private void updateChannel(int channel) {
      this.channel = channel;
      boolean isTransmittedChannel = channel == transmittedChannel;
      int index = channel - 1;
      guvTableModel.setChannel(index, isTransmittedChannel);
      profileTableModel.setChannel(index);
      updateStatistics(index);
   }
}
