package segmentation;

import java.io.File;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Pattern;

/**Helper class to handle reading directories with multifile images.*/
class ImageDirectoryReader {

   private enum Format {

      OME_TIFF(true, ".companion.ome", ".ome.tif", "ome.btf"),
      METAMORPH(true,".nd", ".tif", ".stk"),
      ZEISS(false, null, ".czi", ".lsm");

      private static final EnumSet<Format> set = EnumSet.allOf(Format.class);

      private final String companionExtension;
      private final String[] extensions;
      private final boolean isMultiFile;

      Format(boolean isMultiFile, String companionExtension, String... extensions) {
         this.isMultiFile = isMultiFile;
         this.companionExtension = companionExtension;
         this.extensions = extensions;
      }

      /**Checks if an image file name has a valid extension (mot considering extensions of companion files).
       * @param name the file name with extension
       * @return <code>true</code> if valid, <code>false</code> otherwise*/
      private boolean isFormat(String name) {
         for (String extension : extensions) {
            if (name.endsWith(extension)) {
               return true;
            }
         }
         return false;
      }

      /**Returns the common name part of a multifile image.
       * @param companion the full name of the companion file
       * @return the common name string or an empty string (if the format is not recognized, or if it is not a multi-
       * file format). It never returns <code>null</code>*/
      private static String getCommonNameFromCompanion(String companion) {
         for (Format format : set) {
            if (format.isMultiFile && companion.endsWith(format.companionExtension)) {
                  return companion.substring(0, companion.lastIndexOf(format.companionExtension));
            }
         }
         return "";
      }
   }

   private static final Pattern ND_CHANNEL_PATTERN = Pattern.compile("\"WaveName(\\d)\",\\s(.+)");

   private final ArrayList<File> results;

   /**Constructor.
    * @param directory the image directory*/
   private ImageDirectoryReader(Path directory) {
      results = new ArrayList<>();
      File[] files = directory.toFile().listFiles();
      if (files != null) {
         ArrayList<String> multiFilePrefixes = new ArrayList<>();
         List<File> list = new ArrayList<>(Arrays.asList(files)); //we need a mutable list
         Iterator<File> it = list.iterator();
         while (it.hasNext()) {
            File file = it.next();
            String name = file.getName();
            if (file.isDirectory() || !name.contains(".")) { //ignore files without extension
               it.remove();
            } else if (Format.ZEISS.isFormat(name)) {
               it.remove();
               results.add(file);
            } else {
               String commonPart = Format.getCommonNameFromCompanion(name);
               if (!commonPart.isEmpty()) {
                  //The file is a companion file
                  it.remove();
                  results.add(file);
                  multiFilePrefixes.add(commonPart);
               }
            }
         }
         //Second iteration round to remove non-companion multifile images
         for (File file : list) {
            String name = file.getName();
            boolean isMultiFileImage = false;
            for (String commonPart : multiFilePrefixes) {
               if (name.startsWith(commonPart)) {
                  isMultiFileImage = true;
                  break;
               }
            }
            if (!isMultiFileImage && (Format.OME_TIFF.isFormat(name) || Format.METAMORPH.isFormat(name))) {
               results.add(file);
            }
         }
      }
   }

   /**Reads the transmitted light channel number from a MetaMorph ND or OME-TIFF campanion.ome file.
    * @param companion the companion file
    * @return the channel number starting with 1, or 0 if no transmitted light channel has been found*/
   static int getTransmittedChannel(File companion) {
      int channel = 0;


      return channel;
   }

   /**Returns an array of unique image files. For multifile images, companion file is provided only (other files of
    * these file sets are hidden).
    * @return the results as array*/
   static File[] read(Path directory) {
      ImageDirectoryReader reader = new ImageDirectoryReader(directory);
      return reader.results.toArray(new File[0]);
   }
}
