package segmentation;

import ij.ImagePlus;
import ij.measure.Calibration;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import shared.GUV;

/**Stores infos to split and re-assemble images.*/
class ImageInfo {

   final Path root;
   final String type, title;
   final int nChannels, nSlices, nFrames;
   final int originalWidth, originalHeight, rescaledWidth, rescaledHeight;
   final boolean exists, isMicrons, isHigherDimensional;

   final Calibration calibration;

   Path copyOfOriginal, roiSet;
   ArrayList<Path> files;

   ImageInfo(BatchInfo batchInfo, ImagePlus imp) throws IOException {
      title = imp.getShortTitle();
      nChannels = imp.getNChannels();
      nSlices = imp.getNSlices();
      nFrames = imp.getNFrames();
      calibration = imp.getCalibration();
      String unit = calibration.getUnit();
      isMicrons = unit.equals("micron") || unit.equals("µm") || unit.equals("um");
      originalWidth = imp.getWidth();
      originalHeight = imp.getHeight();
      double pixelWidth = imp.getCalibration().pixelWidth;
      double pixelHeight = imp.getCalibration().pixelHeight;
      rescaledWidth = (int) Math.floor(imp.getWidth() * pixelWidth);
      rescaledHeight = (int) Math.floor(imp.getHeight() * pixelHeight);
      if (imp.getType() == ImagePlus.GRAY8) {
         type = "8-bit";
      } else if (imp.getType() == ImagePlus.GRAY16) {
         type = "16-bit";
      } else if (imp.getType() == ImagePlus.GRAY32) {
         type = "32-bit";
      } else {
         type = "RGB";
      }
      if (nSlices > 1 || nFrames > 1) {
         isHigherDimensional = true;
         root = batchInfo.tmp.resolve(imp.getShortTitle());
      } else {
         isHigherDimensional = false;
         root = batchInfo.tmp;
         files = new ArrayList<>();
      }
      if (batchInfo.dst != null && Files.exists(batchInfo.dst.resolve(title + ".tif")) && Files.exists(
            batchInfo.dst.resolve(title + "_cp_masks.tif"))) {
         exists = true;
      } else {
         exists = false;
         if (isHigherDimensional && Files.notExists(root)) {
            Files.createDirectories(root);
         }
         batchInfo.tmpDirs.add(root);
         if (batchInfo.dst != null) {
            copyOfOriginal = batchInfo.dst.resolve(title+".tif");
            roiSet = batchInfo.dst.resolve(title+".zip");
            imp.setProperty("profileExtension", GUV.profileExt);
         }
      }
      batchInfo.infos.add(this);
   }
}
