package segmentation;

import ij.IJ;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JRadioButton;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.border.EmptyBorder;
import shared.GUV;

/**User interface*/
class ProgressDialog extends JDialog implements ActionListener, PropertyChangeListener {

   private final JProgressBar preprocessBar, cellposeBar, postprocessBar;
   private JRadioButton keepButton, logoutButton, shutdownButton;
   private AbstractProcessor currentProcessor;
   private JProgressBar currentProgressBar;
   private PreProcessor processor;

   ProgressDialog(BatchInfo info, String title) {
      boolean isWindows = System.getProperty("os.name").startsWith("Win");
      String debug = System.getenv("LOGGING");
      if (debug != null && debug.equals("system")) {
         isWindows = true; //for debugging on Linux
      }
      setDefaultCloseOperation(DISPOSE_ON_CLOSE);
      if (isWindows) {
         addWindowListener(new WindowAdapter() {

            @Override
            public void windowClosed(WindowEvent e) {
               if (logoutButton.isSelected()) {
                  shutdown(true);
               } else if (shutdownButton.isSelected()) {
                  shutdown(false);
               }
            }
         });
      }
      setTitle(title);
      Image icon = IJ.getInstance().getIconImage();
      if (icon != null) {
         setIconImage(icon);
      }
      JPanel content = new JPanel();
      content.setLayout(new BoxLayout(content, BoxLayout.PAGE_AXIS));
      content.setBorder(new EmptyBorder(8, 8, 8, 8));
      setContentPane(content);

      JPanel profilePanel = new JPanel(new FlowLayout(FlowLayout.LEADING));
      profilePanel.setAlignmentX(0f);
      profilePanel.add(new JLabel("Profiles"));
      SpinnerNumberModel profileSpinnerModel = new SpinnerNumberModel(GUV.nProfiles,
            4, 30, 1);
      JSpinner profileSpinner = new JSpinner(profileSpinnerModel);
      profileSpinner.addChangeListener(e -> GUV.nProfiles = (int) profileSpinnerModel.getNumber());
      profilePanel.add(profileSpinner);
      content.add(profilePanel);
      JPanel extPanel = new JPanel(new FlowLayout(FlowLayout.LEADING));
      extPanel.setAlignmentX(0f);
      content.add(extPanel);
      extPanel.add(new JLabel("Extend profiles by"), BorderLayout.EAST);
      SpinnerNumberModel extSpinnerModel = new SpinnerNumberModel(GUV.profileExt,
            3, 20, 1);
      JSpinner extSpinner = new JSpinner(extSpinnerModel);
      extSpinner.addChangeListener(e -> GUV.profileExt = (int) extSpinnerModel.getNumber());
      extSpinner.setAlignmentX(0f);
      Dimension d = new Dimension(extSpinner.getPreferredSize().width, extSpinner.getPreferredSize().height);
      extSpinner.setMaximumSize(d);
      extPanel.add(extSpinner);
      extPanel.add(new JLabel("µm"));
      JPanel transmittedPanel = new JPanel(new FlowLayout(FlowLayout.LEADING));
      transmittedPanel.setAlignmentX(0f);
      content.add(transmittedPanel);
      transmittedPanel.add(new JLabel("Transmitted channel"));
      String[] options = {"auto-detect", "none", "channel 1", "channel 2", "channel 3", "channel 4"};
      JComboBox<String> transmittedCombo = new JComboBox<>(options);
      transmittedCombo.setToolTipText("Auto-detection works only with Zeiss .lsm and .czi files");
      transmittedCombo.setSelectedIndex(0);
      transmittedCombo.addActionListener(evt -> {
         JComboBox<String> cb = (JComboBox<String>) evt.getSource();
         String choice = (String) cb.getSelectedItem();
         int channel = 0;
         if (choice != null) {
            if (choice.equals("none")) {
               channel = -1;
            } else if (!choice.equals("auto-detect")) {
               channel = Integer.parseInt(choice.substring(8));
            }
            info.transmittedChannel = channel;
         }
      });
      transmittedPanel.add(transmittedCombo);
      content.add(Box.createVerticalStrut(12));
      JLabel preProcessing = new JLabel("Preprocessing");
      preProcessing.setAlignmentX(0f);
      content.add(preProcessing);
      preprocessBar = new JProgressBar(0, 100);
      preprocessBar.setAlignmentX(0f);
      content.add(preprocessBar);
      JLabel cellpose = new JLabel("Cellpose");
      preProcessing.setAlignmentX(0f);
      content.add(cellpose);
      cellposeBar = new JProgressBar(0, 100);
      cellposeBar.setAlignmentX(0f);
      content.add(cellposeBar);
      JLabel postProcess = new JLabel("Postprocessing");
      postProcess.setAlignmentX(0f);
      content.add(postProcess);
      postprocessBar = new JProgressBar(0, 100);
      postprocessBar.setAlignmentX(0f);
      content.add(postprocessBar);
      if (isWindows) {
         content.add(Box.createVerticalStrut(8));
         ButtonGroup group = new ButtonGroup();
         keepButton = new JRadioButton("Keep PC running when done");
         keepButton.setAlignmentX(0f);
         keepButton.setSelected(true);
         group.add(keepButton);
         content.add(keepButton);
         shutdownButton = new JRadioButton("Shutdown PC when done");
         shutdownButton.setAlignmentX(0f);
         group.add(shutdownButton);
         content.add(shutdownButton);
         logoutButton = new JRadioButton("Logout when done");
         logoutButton.setAlignmentX(0f);
         group.add(logoutButton);
         content.add(logoutButton);
         content.add(Box.createVerticalStrut(4));
      }
      content.add(Box.createVerticalStrut(8));
      JPanel buttonPanel = new JPanel();
      buttonPanel.setLayout(new GridLayout(1, 2));
      buttonPanel.setAlignmentX(0f);
      content.add(buttonPanel);
      JButton startButton = new JButton("Start");
      buttonPanel.add(startButton);
      JButton cancelButton = new JButton("Cancel");
      cancelButton.setEnabled(false);
      cancelButton.addActionListener(this);
      buttonPanel.add(cancelButton);
      startButton.addActionListener(e -> {
         if (processor != null) {
            profileSpinner.setEnabled(false);
            extSpinner.setEnabled(false);
            transmittedCombo.setEnabled(false);
            startButton.setEnabled(false);
            cancelButton.setEnabled(true);
            GUV.savePreferences();
            processor.execute();
         }
      });
      pack();
      setSize(400, getPreferredSize().height);
      setLocationRelativeTo(null);
      setVisible(true);
   }

   @Override
   public void actionPerformed(ActionEvent evt) {
      if (currentProcessor != null) {
         currentProcessor.cancel(true);
      }
      keepButton.setSelected(true);
      dispose();
   }

   @Override
   public void propertyChange(PropertyChangeEvent event) {
      String name = event.getPropertyName();
      if (name.equals("state")) {
         AbstractProcessor processor = (AbstractProcessor) event.getSource();
         String value = event.getNewValue().toString();
         if (value.equals("STARTED")) {
            currentProcessor = processor;
            if (processor instanceof PreProcessor) {
               currentProgressBar = preprocessBar;
            } else if (processor instanceof CellposeProcessor) {
               currentProgressBar = cellposeBar;
            } else if (processor instanceof PostProcessor) {
               currentProgressBar = postprocessBar;
            }
         } else if (value.equals("DONE")) {
            if (processor instanceof PostProcessor) {
               dispose();
            }
         }
      } else if (name.equals("progress")) {
         currentProgressBar.setValue((int) event.getNewValue());
      }
   }

   void setPreprocessor(PreProcessor preProcessor) {
      this.processor = preProcessor;
   }

   /**Shutdown of the computer via a JVM shutdown hook after 10 s. On Windows only.
    * @param logout logout rather than shutdown*/
   private void shutdown(boolean logout) {
      ProcessBuilder builder;
      if (logout) {
         builder = new ProcessBuilder("shutdown", "/l");
      } else {
         builder = new ProcessBuilder("shutdown", "/s", "/t", "10");
      }
      Runtime.getRuntime().addShutdownHook(new Thread(() -> {
         try {
            builder.start();
         } catch (IOException ignored) {}
      }));
      IJ.run("Quit");
   }
}
