package segmentation;

import ij.IJ;
import ij.ImagePlus;
import ij.Macro;
import ij.Prefs;
import ij.WindowManager;
import ij.plugin.PlugIn;
import ij.plugin.frame.Recorder;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import profiler.Profiler;
import shared.GUV;
import shared.IJFileChooser;
import shared.IJLogger;
import shared.MacroOptions;

/**Main class for GUV detection by cellpose*/
public class Segmenter implements PlugIn {

   private static final String TITLE = "Batch Segmentation";
   private static final String PREF_SRC = "guv.profiler.source";
   private static final String PREF_DST = "guv.profiler.destination";

   @Override
   public void run(String arg) {
      IJLogger logger = new IJLogger(TITLE);
      Path path = Paths.get(IJ.getDirectory("ImageJ"), "scripts");
      BatchInfo info = new BatchInfo(path);
      MacroOptions<Arguments> options = null;
      String logging = System.getenv("LOGGING");
      if (logging != null && logging.equals("system")) {
         logger.setSystemLogging();
      }
      Profiler.checkPlugins();
      GUV.readPreferences();
      String args = Macro.getOptions();
      if (args == null || args.isEmpty()) {
         /*We are running interactively in the IJ GUI. If there is an active image, process it. If not, ask
          * for directories.*/
         if (logging == null) {
            logger.setInteractiveLogging();
         }
         if (info.script == null) {
            logger.error("cellpose script missing. It should be in \n"+path);
            return;
         }
         ImagePlus imp = WindowManager.getCurrentImage();
         if (imp == null) {
            String initial = Prefs.get(PREF_SRC, IJ.getDirectory("imagej"));
            info.src = IJFileChooser.getDirectory("Select source...", new File(initial));
            if (info.src == null) {
               return;
            } else if (!Files.isReadable(info.src)) {
               logger.error("Cannot read from source directory "+info.src+".");
               return;
            }
            Prefs.set(PREF_SRC, info.src.toString());
            initial = Prefs.get(PREF_DST, info.src.getParent().toString());
            info.dst = IJFileChooser.getDirectory("Select destination...", new File(initial));
            if (info.dst == null)  {
               return;
            } else if (!Files.isWritable(info.dst)) {
               logger.error("Cannot write in destination directory "+ info.dst +".");
               return;
            }
            Prefs.set(PREF_DST, info.dst.toString());
            info.tmp = info.dst.resolve("tmp");
            if (Files.notExists(info.tmp)) {
               try {
                  Files.createDirectories(info.tmp);
               } catch (IOException e) {
                  logger.error("Cannot create temporary directory in "+info.tmp);
                  return;
               }
            }
         } else {
            try {
               info.tmp = Files.createTempDirectory("cellpose");
            } catch (IOException e) {
               logger.error("Cannot create temporary directory: "+e.getMessage());
               return;
            }
         }
         info.listener = new ProgressDialog(info, TITLE);
      } else {
         //We are running in a macro
         if (logging == null) {
            logger.setSilentLogging();
         }
         if (info.script == null) {
            logger.error("cellpose script missing. It should be in \n "+path);
            return;
         }
         options = new MacroOptions<>(Arguments.class, args, logger);
         info.src = Paths.get(options.getString(Arguments.SOURCE));
         info.dst = Paths.get(options.getString(Arguments.DESTINATION));
         info.tmp = info.dst.resolve("tmp");
         if (Files.notExists(info.src) || Files.notExists(info.dst)) {
            logger.error("Source or destination does not exist.");
            return;
         }
      }
      if (Recorder.record && options != null) {
         StringBuilder sb = new StringBuilder("run(\"");
         sb.append(TITLE);
         sb.append("\", \"");
         sb.append(options.toCommandLine());
         sb.append("\");");
         //Append a newline character (at least on Linux the Recorder does not do this on its own)
         sb.append(System.lineSeparator());
         Recorder.recordString(sb.toString());
      }
      PreProcessor preProcessor = new PreProcessor(logger, info);
      if (info.listener == null) {
         preProcessor.execute();
      } else {
         info.listener.setPreprocessor(preProcessor);
      }
   }
}
