package shared;

import ij.ImagePlus;
import ij.Prefs;
import ij.gui.Line;
import ij.gui.OvalRoi;
import ij.gui.Roi;
import ij.process.ImageStatistics;
import java.awt.Color;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Objects;

/**Container for the coordinates of near-circular ROIs*/
public final class GUV implements Comparable<GUV> {

   /**IJ Preferences setting*/
   public static final String PREF_NPROFILES = "guv.profiler.nProfiles";

   /**IJ Preferences setting*/
   public static final String PREF_PROFILE_EXT = "guv.profiler.profileExt";

   /**Default number of profiles to draw per vesicle*/
   public static final int NPROFILES = 10;

   /**Default number of microns to extend profiles beyond the radius*/
   public static final int PROFILE_EXT = 5;

   /**Current number of microns to extend profiles beyond the maximum radius*/
   public static int profileExt = PROFILE_EXT;

   /**Current number of profiles to draw per vesicle*/
   public static int nProfiles = NPROFILES;

   public final ArrayList<Profile> profiles;
   public final OvalRoi roi;
   final Point2D.Double center;
   final double radius;
   final int nChannels;
   public int id;
   public double[] mean, sd; //average intensity and standard deviation

   private boolean use, isControl;

   /**Constructor for a GUV from a stored oval ROI (with properties describing the GUV).
    * @param roi an oval ROI*/
   public GUV(OvalRoi roi) {
      this.roi = roi;
      id = Integer.parseInt(roi.getProperty("id"));
      use = Boolean.parseBoolean(roi.getProperty("use"));
      isControl = Boolean.parseBoolean(roi.getProperty("control"));
      if (isControl) {
         roi.setStrokeColor(Color.MAGENTA);
      } else {
         roi.setStrokeColor(Color.CYAN);
      }
      nChannels = Integer.parseInt(roi.getProperty("channels"));
      mean = new double[nChannels];
      sd = new double[nChannels];
      ImageStatistics stats = roi.getStatistics();
      center = new Point2D.Double(stats.xCentroid, stats.yCentroid);
      radius = Double.parseDouble(roi.getProperty("radius"));
      profiles = new ArrayList<>();
      for (int i = 0; i < nChannels; i++) {
         mean[i] = Double.parseDouble(roi.getProperty("mean-ch"+i));
         sd[i] = Double.parseDouble(roi.getProperty("SD-ch"+i));
      }
   }

   /**Constructor for a GUV from a new oval ROI.
    * @param id an identifier
    * @param imp the image
    * @param roi the oval ROI
    * @param all a ROI containing all GUVs (may be <code>null</code>*/
   public GUV(int id, ImagePlus imp, OvalRoi roi, int slice, int frame, Roi all) {
      this.id = id;
      this.roi = roi;
      roi.setPosition(0, slice, frame);
      setUsed(true);
      isControl = false;
      nChannels = imp.getNChannels();
      mean = new double[nChannels];
      sd = new double[nChannels];
      roi.setName("G"+id);
      roi.setStrokeColor(Color.CYAN);
      roi.setProperty("control", "false");
      roi.setProperty("id", String.valueOf(id));
      roi.setProperty("channels", String.valueOf(nChannels));
      ImageStatistics stats = roi.getStatistics();
      center = new Point2D.Double(stats.xCentroid, stats.yCentroid);
      Rectangle2D rect = roi.getBounds().getBounds2D();
      int profileExtPixels = (int) (profileExt / imp.getCalibration().pixelWidth);
      radius = Math.min(rect.getWidth(), rect.getHeight())/2 + profileExtPixels;
      roi.setProperty("radius", String.valueOf(radius));
      profiles = new ArrayList<>(nProfiles);
      double theta = 2 * Math.PI / nProfiles;
      //ArrayList<Double> radii = new ArrayList<>();
      for (int i = 0; i < nProfiles; i++) {
         double angle = i * theta;
         double x1 = center.x + radius * (Math.cos(angle));
         double y1 = center.y + radius * (Math.sin(angle));
         if (x1 < 0 || x1 > imp.getWidth() || y1 < 0 || y1 > imp.getHeight()) {
            continue; //don't draw profiles ending outside of the image
         }
         int profileID = i + 1;
         Line line = new Line(center.x, center.y, x1, y1);
         line.setPosition(0, slice, frame);
         line.setName("G"+id+"-L"+profileID);
         line.setProperty("GUV-id", String.valueOf(id));
         line.setProperty("profile-id", String.valueOf(profileID));
         Profile p = new Profile(profileID, line, nChannels);
         if (all != null && all.containsPoint(x1, y1)) {
            p.setUsed(false);
         } else {
            p.setUsed(true);
         }
         profiles.add(p);
      }
   }

   /**Reads parameters from the IJ preferences.<p>
    * <strong>Note: </strong> //Don't use Prefs.getInt() - it does not work.*/
   public static void readPreferences() {
      nProfiles = (int) Prefs.get(PREF_NPROFILES, NPROFILES);
      profileExt = (int) Prefs.get(PREF_PROFILE_EXT, PROFILE_EXT);
   }

   /**Saves parameters to the IJ preferences.*/
   public static void savePreferences() {
      Prefs.set(PREF_NPROFILES, nProfiles);
      Prefs.set(PREF_PROFILE_EXT, profileExt);
      Prefs.savePreferences();
   }

   /**This method should only be used to add a line from a stored ROI set. Don't forget to call the
    * calculate() method once all profiles are added!*/
   public void addProfile(Line line) {
      int profileID = Integer.parseInt(line.getProperty("profile-id"));
      Profile p = new Profile(profileID, line, nChannels);
      p.setUsed(Boolean.parseBoolean(line.getProperty("use")));
      profiles.add(p);
      for (int i = 0; i < nChannels; i++) {
         p.intensity[i] = Double.parseDouble(line.getProperty("intensity-ch"+i));
         p.radius[i] = Double.parseDouble(line.getProperty("radius-ch"+i));
      }
   }

   /**Returns the number of used profiles*/
   public int getNumberOfUsedProfiles() {
      int n = 0;
      for (Profile p : profiles) {
         if (p.isUsed()) {
            n++;
         }
      }
      return n;
   }

   public boolean isControl() {
      return isControl;
   }

   public boolean isUsed() {
      return use;
   }

   /**Measures intensities in the current position of the image provided.
    * @param imp the image*/
   public void measure(ImagePlus imp) {
      for (Profile p : profiles) {
         p.measure(imp);
      }
      calculate(imp.getC() - 1);
   }

   /**Measures intensities in all channels of the image provided. Slices and frames are not changed.*/
   public void measureAll(ImagePlus imp) {
      int currentChannel = imp.getC();
      for (int i = 1; i <= imp.getNChannels(); i++) {
         imp.setC(i);
         measure(imp);
      }
      imp.setC(currentChannel);
   }

   /**Sets this GUV as control/noise reference.
    * @param isControl <code>true</code> or <code>false</code>*/
   public void setAsControl(boolean isControl) {
      this.isControl = isControl;
      roi.setProperty("control", Boolean.toString(isControl));
      if (isControl) {
         roi.setStrokeColor(Color.MAGENTA);
      } else {
         roi.setStrokeColor(Color.CYAN);
      }
   }

   public void setUsed(boolean use) {
      this.use = use;
      roi.setProperty("use", String.valueOf(use));
   }

   /**Comparisons are first made by radius, then by centroid coordinates.*/
   @Override
   public int compareTo(GUV other) {
      if (radius == other.radius) {
         if (center.x == other.center.x) {
            return compare(center.y, other.center.y);
         } else {
            return compare(center.x, other.center.x);
         }
      } else {
         return compare(radius, other.radius);
      }
   }

   @Override
   public boolean equals(final Object obj) {
      if (this == obj) {
         return true;
      } else if (obj == null || obj.getClass() != getClass()) {
         return false;
      }
      GUV other = (GUV) obj;
      if (roi.equals(other.roi)) {
         return true;
      }
      return Objects.equals(center, other.center) && Objects.equals(radius, other.radius);
   }

   @Override
   public int hashCode() {
      return Objects.hash(center, radius);
   }

   /**Calculates statistics for the current stack position.*/
   public void calculate(int channel) {
      double sum = 0;
      int n = 0;
      for (Profile p : profiles) {
         if (p.isUsed()) {
            sum += p.intensity[channel];
            n++;
         }
      }
      if (n == 0) { //user deselected all profiles
         this.use = false;
         mean[channel] = 0;
         sd[channel] = 0;
      } else if (n == 1) { //only one profile, mean valid, standard deviation invalid
         mean[channel] = sum/n;
         sd[channel] = 0;
      } else {
         mean[channel] = sum/n;
         double sumOfSquares = 0;
         for (Profile p : profiles) {
            if (p.isUsed()) {
               sumOfSquares += Math.pow(p.intensity[channel] - mean[channel], 2);
            }
         }
         double meanOfSquares = sumOfSquares/(n - 1);
         sd[channel] = Math.sqrt(meanOfSquares);
      }
      roi.setProperty("mean-ch"+channel, String.valueOf(mean[channel]));
      roi.setProperty("SD-ch"+channel, String.valueOf(sd[channel]));
   }

   /**Calculates statistics for all channels.*/
   public void calculateAll() {
      for (int i = 0; i < nChannels; i++) {
         calculate(i);
      }
   }

//   /**Simple outlier test based on interquartile ranges.
//    * <a href="https://stackoverflow.com/questions/18805178/how-to-detect-outliers-in-an-arraylist">...</a>
//    * @return lower and upper fences*/
//   private Fences findFences(List<Double> values) {
//      Collections.sort(values);
//      List<Double> data1 = values.subList(0, values.size()/2);
//      List<Double> data2;
//      if (values.size() % 2 == 0) {
//         data2 = values.subList(values.size()/2, values.size());
//      } else {
//         data2 = values.subList(values.size()/2 + 1, values.size());
//      }
//      double q1 = getMedian(data1);
//      double q3 = getMedian(data2);
//      double iqr = q3 - q1;
//      return new Fences(q1 - 1.5*iqr, q3 + 1.5*iqr);
//   }

//   private static double getMedian(List<Double> data) {
//      if (data.size() % 2 == 0)
//         return (data.get(data.size()/2) + data.get(data.size()/2 - 1))/2;
//      else
//         return data.get(data.size()/2);
//   }

   private static int compare(double d1, double d2) {
      return Double.compare(d1, d2);
   }
}

///**Helper class to pass fences of interquartile range tests*/
//final class Fences {
//
//   double lower;
//   double upper;
//
//   Fences(double lower, double upper) {
//      this.lower = lower;
//      this.upper = upper;
//   }
//}
