package shared;

import ij.ImagePlus;
import ij.gui.Roi;
import java.util.ArrayList;
import java.util.List;

/**A container for GUVs and auxiliary data.*/
public class GUVSet {

   public final int frame, slice;
   public Roi all;
   public double[] mean, sd;

   private int controlIndex; //index of the background/noise GUV (-1 for none)
   private final ArrayList<GUV> guvs;
   private int nextID;

   /**Creates a new GUV set.
    * @param frame the frame position (time point)
    * @param slice the slice position (z slice)
    * @param nChannels the number of the channels*/
   public GUVSet(int frame, int slice, int nChannels) {
      this.frame = frame;
      this.slice = slice;
      this.guvs = new ArrayList<>();
      mean = new double[nChannels];
      sd = new double[nChannels];
      controlIndex = -1;
      nextID = 1;
   }

   public GUVSet(int frame, int slice, int nChannels, Roi all) {
      this(frame, slice, nChannels);
      this.all = all;
   }

   public void add(GUV guv) {
      guvs.add(guv);
      if (guv.isControl()) {
         controlIndex = guvs.size() - 1;
      }
      if (guv.id >= nextID) {
         nextID = guv.id + 1;
      }
   }

   /**Re-calculates the statistics of this GUV set over all channels.*/
   public void calculateAll() {
      for (int i = 0; i < mean.length; i++) {
         calculate(i);
      }
   }

   /**Re-calculates the statistics of this GUV set.
    * @param channel the channel with 0 being the first (different than {@link ImagePlus#getChannel()}*/
   public void calculate(int channel) {
      double sum = 0;
      int n = 0;
      for (GUV guv : guvs) {
         if (guv.isUsed() && !guv.isControl()) {
            sum += guv.mean[channel];
            n++;
         }
      }
      if (n == 0) { //all GUVs deselected or deleted
         mean[channel] = 0;
         sd[channel] = 0;
      } else if (n == 1) {
         mean[channel] = sum/n;
         sd[channel] = 0;
      } else {
         mean[channel] = sum/n;
         double sumOfSquares = 0;
         for (GUV guv : guvs) {
            if (guv.isUsed() && !guv.isControl()) {
               sumOfSquares += Math.pow(guv.mean[channel] - mean[channel], 2);
            }
         }
         double meanOfSquares = sumOfSquares/(n - 1);
         sd[channel] = Math.sqrt(meanOfSquares);
      }
   }

   public void clear() {
      guvs.clear();
      for (int i = 0; i < mean.length; i++) {
         mean[i] = 0;
         sd[i] = 0;
      }
   }

   public int count() {
      return guvs.size();
   }

   /**Removes the GUVs with the provided index.
    * @param index the index of the GUV
    * @return the removed GUV*/
   public GUV delete(int index) {
      if (index == controlIndex) {
         controlIndex = -1;
      } else if (index >= 0 && index < controlIndex) {
         controlIndex -= 1;
      }
      return guvs.remove(index);
   }

   public GUV get(int index) {
      return guvs.get(index);
   }

   /**Provides the current background GUV
    * @return the GUV, or <code>null</code> if there is currently none*/
   public GUV getControl() {
      if (controlIndex == -1) {
         return null;
      } else {
         return guvs.get(controlIndex);
      }
   }

   /**Provides the index of the current background GUV
    * @return the index*/
   public int getControlIndex() {
      return controlIndex;
   }

   public List<GUV> getAll() {
      return guvs;
   }

   /**Provides the highest GUV ID of this set.
    * @return the ID (not necessarily identical with the GUV count)*/
   public int getNextID() {
      return nextID;
   }

   /**Measures the image at its current position/channel.
    * @param imp the image*/
   public void measure(ImagePlus imp) {
      for (GUV guv : guvs) {
         guv.measure(imp);
      }
      calculate(imp.getChannel() - 1);
   }

   /**Measures the image in all channels at its current frame and slice.
    * @param imp the image*/
   public void measureAllChannels(ImagePlus imp) {
      int channel = imp.getC();
      for (int i = 1; i <= imp.getNChannels(); i++) {
         imp.setC(i);
         measure(imp);
      }
      imp.setC(channel);
   }

   /**Sets the GUV with the provided index as background
    * @param index the index*/
   public void setAsControl(int index) {
      if (controlIndex > -1 && controlIndex < guvs.size()) {
         guvs.get(controlIndex).setAsControl(false);
      }
      if (index < guvs.size() && index > -1) {
         guvs.get(index).setAsControl(true);
      }
      this.controlIndex = index;
   }
}