package shared;

import ij.IJ;
import java.awt.Component;
import java.awt.Font;
import java.awt.Insets;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.UIManager;
import javax.swing.WindowConstants;

/**Very basic logger and messenger.*/
public class IJLogger {

   private String title;
   private Mode mode;
   private Path logPath;

   /**Constructor with default mode (interactive).*/
   @SuppressWarnings("unused")
   public IJLogger() {
      mode = Mode.INTERACTIVE;
   }

   /**Constructor with default mode (interactive).
    * @param title used for dialog titles and log entries*/
   @SuppressWarnings("unused")
   public IJLogger(String title) {
      this();
      setTitle(title);
   }

   /**Activates ImageJ style logging: info and warnings are printed in IJ's log window, error messages are
    * shown in IJ dialogs, exceptions are delegated to IJ.*/
   @SuppressWarnings("unused")
   public void setImageJLogging() {
      this.mode = Mode.IMAGEJ;
   }

   /**Returns the title used for dialogs and log entries.
    * @return the title*/
   @SuppressWarnings("unused")
   public String getTitle() {
      return title;
   }

   /**Sets the title used for dialogs and log entries.
    * @param title the title*/
   @SuppressWarnings("unused")
   public void setTitle(String title) {
      this.title = title;
   }

   /**Activates interactive logging. Warnings, error messages and exceptions and warnings are displayed in
    * dialogs, info messages are printed in to IJ's log window.*/
   @SuppressWarnings("unused")
   public void setInteractiveLogging() {
      this.mode = Mode.INTERACTIVE;
   }

   /**Returns true if the logging mode is interactive.
    * @return <code>true</code> if the logger is in interactive mode*/
   @SuppressWarnings("unused")
   public boolean isInteractiveLogging() {
      return this.mode == Mode.INTERACTIVE;
   }

   /**Activates the silent logging mode. All messages are redirected to IJ's log window.*/
   @SuppressWarnings("unused")
   public void setSilentLogging() {
      this.mode = Mode.SILENT;
   }

   /**Activates file logging. All messages are written to a log file.
    * @param path the log file
    * @return <code>true</code> if the path is writable, <code>false</code> otherwise*/
   @SuppressWarnings("unused")
   public boolean setFileLogging(Path path) {
      boolean success;
      if (Files.notExists(path)) {
         try {
            Files.createFile(path);
         } catch (IOException e) {
            return false;
         }
         success = true;
      } else {
         success = Files.isWritable(path);
      }
      if (success) {
         this.logPath = path;
         this.mode = Mode.FILE;
      }
      return success;
   }

   /**Activates the system logging mode. Infos and warnings are redirected to System.out, errors and
    * exceptions to System.err.*/
   @SuppressWarnings("unused")
   public void setSystemLogging() {
      this.mode = Mode.SYSTEM;
   }

   /**Logs infos.
    * @param message the message*/
   @SuppressWarnings("unused")
   public void info(String message) {
      mode.info(title, message, logPath);
   }

   /**Logs warnings.
    * @param message the warning*/
   @SuppressWarnings("unused")
   public void warning(String message) {
      mode.warning(title, message, logPath);
   }

   /**Logs errors.
    * @param message the message*/
   @SuppressWarnings("unused")
   public void error(String message) {
      mode.error(title, message, logPath);
   }

   /**Logs exceptions.
    * @param e the exception*/
   @SuppressWarnings("unused")
   public void error(Exception e) {
      mode.error(title, e, logPath);
   }

   /**Attempts to delete the current log file if it has remained empty.*/
   @SuppressWarnings("unused")
   public void cleanup() {
      if (logPath != null && Files.isWritable(logPath)) {
         try (BufferedReader br = new BufferedReader(new FileReader(logPath.toFile()))) {
            if (br.readLine() == null) {
               br.close();
               Files.delete(logPath);
            }
         } catch (Exception e) {
            //ignored
         }
      }
   }

   /**Log levels implemented as {@link Enum}.*/
   public enum Mode implements MacroOption<Mode> {

      /**Interactive logging. Infos are printed to ImageJ's log window, all other messages are displayed
       * in dialogs (a bit nicer than ImageJ's ancient AWT dialogs).*/
      @SuppressWarnings("unused")
      INTERACTIVE("interactive", "Interactive", null, "Displays warnings" +
            "and errors in dialogs, prints infos to ImageJ's log window") {

         void info(String title, String message, Path path) {
            IJ.log(title+" [info] "+message);
         }

         void warning(String title, String message, Path path) {
            JPanel panel = new JPanel();
            panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
            Font font = UIManager.getDefaults().getFont("Label.font");
            font = font.deriveFont(Font.BOLD, font.getSize2D() + 1);
            JLabel header = new JLabel("Warning");
            header.setFont(font);
            panel.add(Box.createVerticalStrut(5));
            panel.add(header, Component.LEFT_ALIGNMENT);
            panel.add(Box.createVerticalStrut(10));
            panel.add(new JLabel(message), Component.LEFT_ALIGNMENT);
            panel.add(Box.createVerticalStrut(5));
            JOptionPane.showMessageDialog(null, panel, title,
                  JOptionPane.WARNING_MESSAGE);
         }

         void error(String title, String message, Path path) {
            JPanel panel = new JPanel();
            panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
            Font font = UIManager.getDefaults().getFont("Label.font");
            font = font.deriveFont(Font.BOLD, font.getSize2D() + 1);
            JLabel header = new JLabel("Error");
            header.setFont(font);
            panel.add(Box.createVerticalStrut(5));
            panel.add(header, Component.LEFT_ALIGNMENT);
            panel.add(Box.createVerticalStrut(10));
            panel.add(new JLabel(message), Component.LEFT_ALIGNMENT);
            panel.add(Box.createVerticalStrut(5));
            JOptionPane.showMessageDialog(null, panel, title,
                  JOptionPane.ERROR_MESSAGE);
         }

         void error(String title, Exception e, Path path) {
            JPanel panel = new JPanel();
            panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
            Font font = UIManager.getDefaults().getFont("Label.font");
            font = font.deriveFont(Font.BOLD, font.getSize2D() + 1);
            JLabel header = new JLabel("Exception");
            header.setFont(font);
            panel.add(Box.createVerticalStrut(5));
            panel.add(header, Component.LEFT_ALIGNMENT);
            panel.add(Box.createVerticalStrut(10));
            panel.add(new JLabel(e.getMessage()), Component.LEFT_ALIGNMENT);
            panel.add(Box.createVerticalStrut(5));
            panel.add(new JLabel(e.getStackTrace()[0].toString()),
                  Component.LEFT_ALIGNMENT);
            panel.add(Box.createVerticalStrut(10));
            StringWriter stackTraceWriter = new StringWriter();
            e.printStackTrace(new PrintWriter(stackTraceWriter));
            JTextArea textArea = new JTextArea(stackTraceWriter.toString(), 8, 40);
            textArea.setMargin(new Insets(2, 2, 2, 2));
            JScrollPane scroll = new JScrollPane(textArea);
            scroll.setAlignmentX(0f);
            panel.add(scroll);
            String[] options = {"OK", "Save..."};
            JOptionPane optionPane = new JOptionPane(panel,
                  JOptionPane.ERROR_MESSAGE, JOptionPane.YES_NO_OPTION, null, options, options[0]);
            JDialog dialog = optionPane.createDialog(null, title+" - Error");
            dialog.setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
            dialog.addWindowListener(new WindowAdapter() {

               public void windowClosing(WindowEvent we) {
                  dialog.dispose();
               }
            });
            optionPane.addPropertyChangeListener(evt -> {
               String propertyName = evt.getPropertyName();
               if (dialog.isVisible() && (evt.getSource() == optionPane)
                     && (propertyName.equals(JOptionPane.VALUE_PROPERTY))) {
                  dialog.setVisible(false);
               }
            });
            dialog.setResizable(true);
            dialog.pack();
            dialog.setVisible(true);
            if (optionPane.getValue().equals("Save...")) {
               Path save = Paths.get(IJ.getDirectory("current"), title+".txt");
               File f = IJFileChooser.saveFile(title, save.toFile(), "Text files",
                     new String[] {"txt", "log"});
               if (f != null) {
                  if (f.exists()) {
                     int value = JOptionPane.showConfirmDialog(null, f.getName()
                           +" exists. Overwrite?", title, JOptionPane.YES_NO_OPTION,
                           JOptionPane.QUESTION_MESSAGE);
                     if (value == JFileChooser.APPROVE_OPTION) {
                        StringBuilder sb = new StringBuilder();
                        sb.append("Exception in: ");
                        sb.append(title);
                        sb.append(System.lineSeparator());
                        sb.append("Message: ");
                        sb.append(e.getMessage());
                        sb.append(System.lineSeparator());
                        sb.append("Caused by: ");
                        sb.append(e.getStackTrace()[0].toString());
                        sb.append(System.lineSeparator());
                        sb.append(System.lineSeparator());
                        sb.append("Full stack trace:");
                        sb.append(System.lineSeparator());
                        sb.append(stackTraceWriter);
                        try {
                           Files.write(Paths.get(f.toURI()), sb.toString().getBytes(StandardCharsets.UTF_8),
                                 StandardOpenOption.CREATE,
                                 StandardOpenOption.TRUNCATE_EXISTING);
                        } catch (IOException ioe) {
                           //ignored
                        }
                     }
                  }
               }
            }
         }
      },

      /**ImageJ1 style logging. Infos and warnings are printed to IJ's log window, errors
       * are shown in an IJ dialog, exceptions are handled by IJ.*/
      @SuppressWarnings("unused")
      IMAGEJ("imagej", "ImageJ", null, "Delegates logging to ImageJ") {

         void info(String title, String message, Path path) {
            IJ.log(title+" [INFO] "+message);
         }

         void warning(String title, String message, Path path) {
            IJ.log(title+" [WARNING] "+message);
         }

         void error(String title, String message, Path path) {
            IJ.error(title+" [ERROR] ", message);
         }

         void error(String title, Exception e, Path path) {
            IJ.handleException(e);
         }
      },

      /**Silent logging. No dialogs are displayed. All messages are printed to ImageJ's log window.*/
      @SuppressWarnings("unused")
      SILENT("silent", "Silent", null, "Prints all messages to ImageJ's" +
            "log window") {

         void info(String title, String message, Path path) {
            IJ.log(title+" [INFO] "+message);
         }

         void warning(String title, String message, Path path) {
            IJ.log(title+" [WARNING] "+message);
         }

         void error(String title, String message, Path path) {
            IJ.log(title+" [ERROR] "+message);
         }

         void error(String title, Exception e, Path path) {
            IJ.log(title+" [ERROR] "+e.getClass().getSimpleName()+" in "
                  +e.getStackTrace()[0].toString());
            IJ.log("      "+e.getMessage());
         }
      },

      /**All messages are written to a log file.*/
      @SuppressWarnings("unused")
      FILE("file", "Log file", Path.class, "Writes all messages to a log file") {

         private final DateTimeFormatter FORMATTER = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss");

         void info(String title, String message, Path path) {
            write(format(title, "info", message), path);
         }

         void warning(String title, String message, Path path) {
            write(format(title, "warning", message), path);
         }

         void error(String title, String message, Path path) {
            write(format(title, "error", message), path);
         }

         void error(String title, Exception e, Path path) {
            String message = e.getClass().getSimpleName()+" "+e.getStackTrace()[0].toString();
            StringWriter stackTraceWriter = new StringWriter();
            e.printStackTrace(new PrintWriter(stackTraceWriter));
            write(stackTraceWriter.toString(), path);
         }

         private String format(String title, String type, String message) {
            StringBuilder sb = new StringBuilder();
            sb.append(LocalDateTime.now().format(FORMATTER));
            sb.append(" ");
            sb.append(title);
            sb.append(" [");
            sb.append(type);
            sb.append("] ");
            sb.append(message);
            sb.append(System.lineSeparator());
            return sb.toString();
         }

         private void write(String message, Path path) {
            try {
               Files.write(path, message.getBytes(StandardCharsets.UTF_8), StandardOpenOption.APPEND);
            } catch (Exception e) {
               //ignore
            }
         }
      },

      /**Infos and warnings are redirected to the System.out, errors and exceptions to System.err.*/
      @SuppressWarnings("unused")
      SYSTEM("system", "System", null, "Redirects all messages to" +
            "System.out/System.err") {

         void info(String title, String message, Path path) {
            System.out.println(title+" [INFO] "+message);
         }

         void warning(String title, String message, Path path) {
            System.out.println(title+" [WARNING] "+message);
         }

         void error(String title, String message, Path path) {
            System.err.println(title+" [ERROR] "+message);
         }

         void error(String title, Exception e, Path path) {
            System.err.println(title+" [ERROR] "+e.getClass().getSimpleName()+" in "
                  +e.getStackTrace()[0].toString());
            e.printStackTrace();
         }
      };

      final String command, label, description;
      final Class<?> parameterClass;

      /**Constructor.
       * @param command a scriptable command
       * @param label a label for UI elements
       * @param parameterClass the parameter class
       * @param description the short description for UI elements*/
      Mode(String command, String label, Class<?> parameterClass, String description) {
         this.command = command;
         this.label = label;
         this.parameterClass = parameterClass;
         this.description = description;
      }

      /**{@inheritDoc}*/
      @SuppressWarnings("unused")
      @Override
      public String getCommand() {
         return command;
      }

      /**{@inheritDoc}*/
      @SuppressWarnings("unused")
      @Override
      public String getLabel() {
         return label;
      }

      /**{@inheritDoc}*/
      @SuppressWarnings("unused")
      public Class<?> getParameterClass() {
         return parameterClass;
      }

      /**{@inheritDoc}*/
      @SuppressWarnings("unused")
      public String getDescription() {
         return description;
      }

      /**Logs infos.
       * @param title for dialog titles and log entry headers
       * @param message the message
       * @param path path to a log file*/
      @SuppressWarnings("unused")
      abstract void info(String title, String message, Path path);

      /**Logs warnings.
       * @param title for dialog titles and log entry headers
       * @param message the warning
       * @param path path to a log file*/
      @SuppressWarnings("unused")
      abstract void warning(String title, String message, Path path);

      /**Logs errors.
       * @param title for dialog titles and log entry headers
       * @param message the message
       * @param path path to a log file*/
      @SuppressWarnings("unused")
      abstract void error(String title, String message, Path path);

      /**Logs exceptions.
       * @param title for dialog titles and log entry headers
       * @param e the exception*/
      @SuppressWarnings("unused")
      abstract void error(String title, Exception e, Path path);
   }
}