package shared;

import java.util.Arrays;
import java.util.Collections;
import java.util.EnumMap;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**Utility class to handle macro options implemented as {@link Enum}, forced to implement common methods
 * by the {@link MacroOption} interface.*/
public class MacroOptions<T extends Enum<T> & MacroOption<T>> {

   /**Matches strings with any characters between square brackets.*/
   @SuppressWarnings("RegExpRedundantEscape")
   private static final Pattern BRACKET_PATTERN = Pattern.compile("(\\[.+\\])");

   /**Arbitrary string to temporarily replace spaces in parameters protected by brackets.*/
   private static final String protection = "###";

   private final Class<T> c;
   private final Set<T> all;
   private EnumMap<T, Object> map;
   private IJLogger log;

   /**Default constructor.
    * @param options the {@link Enum}, which must also implement the {@link MacroOption} interface*/
   @SuppressWarnings("unused")
   public MacroOptions(Class<T> options) {
      this.c = options;
      this.all = Collections.unmodifiableSet(EnumSet.allOf(c));
   }

   /**Constructor. Options should follow ImageJ style, with commands (keys) followed by an equals sign and
    * parameters (values). Parameters containing spaces must be protected by square brackets. If
    * commands are not followed by an equals sign and parameter, they are treated as boolean commands
    * with parameter <code>true</code> if present or <code>false</code> if absent:<br>
    *    <br>
    *       <code>arg1=doSomething arg2=[C:\\Program Files\\Java] arg3</code><br>
    *    <br>
    * @param options the {@link Enum}, which must also implement the {@link MacroOption} interface
    * @param arguments the command line passed on by macros
    * @param log the logger*/
   @SuppressWarnings("unused")
   public MacroOptions(Class<T> options, String arguments, IJLogger log) {
      this(options);
      this.map = new EnumMap<>(options);
      this.log = log;

      //Argument parsing
      Matcher m = BRACKET_PATTERN.matcher(arguments);
      String protectedArguments = arguments;
      while (m.find()) {
         String str = m.group(1).replace(" ", protection);
         protectedArguments = protectedArguments.replace(m.group(1), str);
      }
      for (String arg : protectedArguments.split(" ")) {
         //noinspection RegExpRedundantEscape
         arg = arg.replaceAll(protection, " ").replaceAll("\\[(.*)\\]", "$1");
         String[] split = arg.split("=");
         T option = getForCommand(split[0].trim());
         if (option == null) {
            log.warning("Ignoring unknown marco argument '"+split[0]+"'");
            continue;
         }
         if (split.length == 1 && option.getParameterClass() == Boolean.TYPE) {
            map.put(option, Boolean.TRUE);
         } else if (split.length == 2) {
            Object parameter = parseParameter(option, split[1].trim());
            if (parameter != null) {
               map.put(option, parameter);
            }
         } else {
            log.warning("Ignoring malformed command '"+arg+"'");
         }
      }
   }

   /**Returns <code>true</code> if the option supplied as argument is present in the macro options,
    * <code>false</code> otherwise.
    * @param option the option
    * @return <code>true</code> or <code>false</code>*/
   @SuppressWarnings("unused")
   public boolean contains(T option) {
      return map != null && map.containsKey(option);
   }

   /**Returns all options as unmodifiable set.
    * @return the set*/
   @SuppressWarnings("unused")
   public Set<T> getAll() {
      return all;
   }

   /**Returns all options as array.
    * @return the array*/
   @SuppressWarnings("unused")
   public T[] getAllAsArray() {
      return c.getEnumConstants();
   }

   /**Returns the set of all possible commands.
    * @return the set (never <code>null</code>)*/
   @SuppressWarnings("unused")
   public Set<String> getCommands() {
      Set<String> set = new HashSet<>(all.size());
      for (T option : all) {
         set.add(option.getCommand());
      }
      return set;
   }

   /**Returns the option represented by a macro command.
    * @param command the macro command
    * @return the option or {@code null} if there is no matching option*/
   @SuppressWarnings("unused")
   public T getForCommand(String command) {
      for (T option : all) {
         if (option.getCommand().equals(command)) {
            return option;
         }
      }
      return null;
   }

   /**Returns the option represented by a label.
    * @param label the label of the option
    * @return the option or {@code null} if there is no matching option*/
   @SuppressWarnings("unused")
   public T getForLabel(String label) {
      for (T option : all) {
         if (option.getLabel().equals(label)) {
            return option;
         }
      }
      return null;
   }

   /**Returns the parameter of the supplied option.
    * @param option the option
    * @return the parameter or, if the option has not been set, <code>null</code>*/
   @SuppressWarnings("unused")
   public Object getParameter(T option) {
      return map == null ? null : map.get(option);
   }

   /**Returns the parameter of the supplied option as string.
    * @param option the option
    * @return the parameter. Returns <code>null</code> if the option is not set*/
   @SuppressWarnings("unused")
   public String getString(T option) {
      if (map == null) {
         return null;
      }
      Object value = map.get(option);
      return value == null ? null : value.toString();
   }

   /**Returns the parameter of the supplied option as boolean.
    * @param option the option
    * @return <code>true</code> or <code>false</code>. Returns <code>false</code> if the option has
    * not been set, <code>null</code> if the option cannot be parsed*/
   @SuppressWarnings("unused")
   public Boolean getBoolean(T option) {
      if (map == null || option.getParameterClass() != Boolean.TYPE) {
         return null;
      }
      if (!map.containsKey(option)) {
         return false;
      }
      Object value = map.get(option);
      if (!(value instanceof Boolean)) {
         return null;
      }
      return (Boolean) value;
   }

   /**Returns the parameter of the supplied option as integer.
    * @param option the option
    * @return the parameter. Returns <code>null</code> if the option is not set or cannot be parsed*/
   @SuppressWarnings("unused")
   public Integer getInteger(T option) {
      if (map == null || option.getParameterClass() != Integer.TYPE) {
         return null;
      }
      Object value = map.get(option);
      if (!(value instanceof Integer)) {
         return null;
      }
      return (Integer) value;
   }

   /**Returns the parameter of the supplied option as double.
    * @param option the option
    * @return the parameter. Returns <code>null</code> if the option is not set or cannot be parsed*/
   @SuppressWarnings("unused")
   public Double getDouble(T option) {
      if (map == null || option.getParameterClass() != Double.TYPE) {
         return null;
      }
      Object value = map.get(option);
      if (!(value instanceof Double)) {
         return null;
      }
      return (Double) value;
   }

   /**Returns the command line resulting in the current set of arguments*/
   @SuppressWarnings("unused")
   public String toCommandLine() {
      StringBuilder sb = new StringBuilder();
      if (map != null && !map.isEmpty()) {
         for (T option : map.keySet()) {
            Class<?> c = option.getParameterClass();
            sb.append(option.getCommand());
            if (c != Boolean.TYPE) {
               sb.append("=");
               String param;
               if (c.isEnum()) {
                  @SuppressWarnings("rawtypes")
                  MacroOption m = (MacroOption) map.get(option);
                  param = m.getCommand();
               } else {
                  param = map.get(option).toString();
               }
               boolean hasSpaces = false;
               if (c == String.class && param.contains(" ")) {
                  hasSpaces = true;
                  sb.append("[");
               }
               sb.append(param);
               if (hasSpaces) {
                  sb.append("]");
               }
            }
            sb.append(" ");
         }
         sb.delete(sb.length() - 1, sb.length());
      }
      return sb.toString();
   }

   @Override
   public String toString() {
      StringBuilder sb = new StringBuilder("<");
      sb.append(c.getSimpleName());
      sb.append(">");
      if (map != null && !map.isEmpty()) {
         sb.append("{");
         for (T option : map.keySet()) {
            sb.append(option.name());
            sb.append("=");
            sb.append(map.get(option));
            sb.append(", ");
         }
         sb.delete(sb.length() - 2, sb.length());
         sb.append("}");
      }
      return sb.toString();
   }

   /**Tries to parse a string into an object of the class specified.
    * @param option the option
    * @param str the parameter
    * @return the parsed object or <code>null</code> if parsing failed*/
   @SuppressWarnings("rawtypes")
   private Object parseParameter(T option, String str) {
      Class<?> c = option.getParameterClass();
      if (c == String.class) {
         return str;
      } else if (c == Boolean.TYPE) {
         return Boolean.parseBoolean(str);
      } else if (c.isEnum() && Arrays.asList(c.getInterfaces()).contains(MacroOption.class)) {
         Enum[] enums = (Enum[]) c.getEnumConstants();
            for (Enum e : enums) {
               MacroOption m = (MacroOption) e;
               if (m.getCommand().equals(str)) {
                  return m;
               }
            }
      }
      try {
         if (c == Integer.TYPE) {
            return Integer.parseInt(str);
         }  else if (c == Double.TYPE) {
            return Double.parseDouble(str);
         }
      } catch (NumberFormatException nfe) {
         log.warning("Cannot convert '"+str+"' into a number");
         return null;
      }
      log.warning("Unsupported parameter class '"+c.getSimpleName()+"' in option "+ option +
            ". Will store value as object");
      return str;
   }
}
